#define _OMF2D_VERSION 0x101

#define WIN32_LEAN_AND_MEAN
//#define UNICODE
#ifdef UNICODE
  #define _UNICODE
#endif
#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include <tchar.h>

#define OMF_COMENT  0x88
#define OMF_EXTDEF  0x8C
#define OMF_PUBDEF  0x90
#define OMF_PUBD32  0x91
#define OMF_LNAMES  0x96
#define OMF_SEGDEF  0x98
#define OMF_SEGD32  0x99
#define OMF_FIXUPP  0x9C
#define OMF_FIXU32  0x9D
#define OMF_LEDATA  0xA0
#define OMF_LEDA32  0xA1
#define OMF_LIDATA  0xA2
#define OMF_LIDA32  0xA3
#define OMF_COMDEF  0xB0
#define OMF_LEXTDEF 0xB4
#define OMF_LEXTD32 0xB5
#define OMF_LPUBDEF 0xB6
#define OMF_LPUBD32 0xB7
#define OMF_LCOMDEF 0xB8
#define OMF_CEXTDEF 0xBC
#define OMF_COMDAT  0xC2
#define OMF_COMD32  0xC3
#define OMF_LINSYM  0xC4
#define OMF_LINS32  0xC5
#define OMF_LLNAMES 0xCA

#define MAX_SEGMENTS       256
#define MAX_COMDEFSEGMENTS 8192
#define MAX_EXTDEFS        8192
#define MAX_LNAMES         8192
#define FILESIZE_PLUS      65536
#define OMF2D_MAX_LNAMES_TO_CHANGE  128
#define OMF2D_MAX_ENAMES_TO_CHANGE  512
#define OMF2D_MAX_PNAMES_TO_CHANGE 1024


#define OMF2D_ORDINALANY   0
typedef BYTE (WINAPI *TUndecorateName)(PBYTE OrgName, PBYTE NewName, BOOL KeepLeading_, BOOL KeepEnd_);
BYTE WINAPI DefaultUndecorateName(PBYTE OrgName, PBYTE NewName, BOOL KeepLeading_, BOOL KeepEnd_);
TUndecorateName UndecorateName = DefaultUndecorateName;

#define OMF2D_ED_DLLIMPORT 1
#define OMF2D_ED_ISABS     2

#pragma pack(1)

static PBYTE pwrite;

static BOOL  gNameEditor          = FALSE;
static BOOL  DoNotHandleDllImport = FALSE;
static BOOL  DoNotRemoveFixups    = FALSE;
static BOOL  LocalExtPubName      = FALSE;
static BOOL  DoNotUndecorate      = FALSE;
static BOOL  KeepLeading_         = FALSE;
static BOOL  KeepEnd_             = FALSE;

typedef struct _OldNewName {
  WORD  CurOrdinal;
  WORD  Ordinal;
  PBYTE OldName;
  PBYTE NewName;
} OldNewName, *POldNewName;
OldNewName NewLNames[OMF2D_MAX_LNAMES_TO_CHANGE];
OldNewName NewENames[OMF2D_MAX_ENAMES_TO_CHANGE];
OldNewName NewPNames[OMF2D_MAX_PNAMES_TO_CHANGE];
DWORD  NewLNamesCounter = 0, NewENamesCounter = 0, NewPNamesCounter = 0;

typedef struct _ComdefSegment {
   WORD RealSegmentIndex;
  DWORD Start;
  DWORD PrivateOffset;
} ComdefSegment, *PComdefSegment;
ComdefSegment ComdefSegments[MAX_COMDEFSEGMENTS];


typedef struct _ExtDefValue {
  PBYTE ExtDefName;
  DWORD dwValue;
} ExtDefValue, *PExtDefValue;
ExtDefValue NewEDValues[MAX_EXTDEFS];
DWORD  NewEDValuesCounter = 0;

typedef struct _ExtDef {
  BYTE  dbFlags;
  DWORD dwValue;
} ExtDef, *PExtDef;

#include "OMF2D_UndecorateName.h"
#include "OMF2D_ChangeName.h"
#include "OMF2D_GetLBS.h"
#include "OMF2D_GetSetIndex.h"
#include "OMF2D_GetOptions.h"

int _tmain(int argc, TCHAR **argv) {
    DWORD  RecordLength;
    BYTE   RecordType;
    DWORD  SEGDXXCounter = 0;
    DWORD  ExtdefCounter = 0;
    DWORD  LNamesCounter = 0;

    DWORD  OrgNameLength, NewNameLength, i, j;
    DWORD  GroupIndex, SegmentIndex, TypeIndex, NameIndex;
    DWORD  GroupIndexLength, SegmentIndexLength, TypeIndexLength, NameIndexLength;
    DWORD  _TEXTIndexLength = 1, _TEXTIndex = 1, _DATAIndexLength = 1, _DATAIndex = 2; //FIX later
    DWORD  ComSize;
    DWORD  ComdefSegmentIndex, FrameComdefSegmentIndex, TargetComdefSegmentIndex;
    PBYTE  precend, prstring, pwstring;
    PBYTE  prFixupStart, pwFixupStart;
    DWORD  FrameDatum, TargetDatum;
    BOOL   RemoveThisFIXU32, SkipThisFIXU32, NameEditor;
    BYTE   OrgByte[2], NewByte[2], M;

    DWORD  DataOffset, TwoOrFour, LastLIDAXXSize;
    PBYTE  pLastLXDAXX, pLastLXDAXXRecord;
    PBYTE  pPreLastLXDAXX = NULL;
    DWORD  LastLXDAXXSize, PreLastLXDAXXSize;
    DWORD  LXDAXXOffsets[MAX_SEGMENTS] = {0};

    PDWORD pSegmentSize[MAX_SEGMENTS];
    DWORD  ComdatOffsets[MAX_SEGMENTS];
    ExtDef ExtDefs[MAX_EXTDEFS] = {{0, 0}};
    PBYTE  LNames[MAX_LNAMES];

/////////////////////////////////
  __try {

    _ftprintf(stderr,
      _T("\nOMF2D %u.%02u converts 32bit OMF to Delphi linkable OMF")
      _T("\nCopyright (C) 2003 Radim Picha, www.anticracking.sk/EliCZ\n"),
      HIBYTE(_OMF2D_VERSION), LOBYTE(_OMF2D_VERSION));
    if(argc-- < 2)
      return(_ftprintf(stderr,
               _T("Usage: OMF2D <InputOMF> [OutputOMF] [Options]")
               _T("\nOptions:")
               _T("\n/C+   Change names only")
               _T("\n/CL[#]name=name   Change [#th occurrence of] LNAME")
               _T("\n/CE[#]name=name   Change [#th occurrence of] EXT/COM/EXPDEF name")
               _T("\n/CP[#]name=name   Change [#th occurrence of] PUBDEF name")
               _T("\n/U-   Do not undecorate EXT/COM/EXP/PUBDEF names")
               _T("\n/UUdll   dll exporting user's UndecorateName")
               _T("\n/U_*   Keep leading underscore of EXT/COM/EXP/PUBDEF names")
               _T("\n/U*_   Keep end     underscore of EXT/COM/EXP/PUBDEF names")
               _T("\n/D-   Do not handle __declspec(dllimport)")
               _T("\n/SEname=value   Take EXTDEF as ABS")
               _T("\n/F+   Keep non Offset32 (debug) fixups")
               _T("\n/Ofile   Read options from file\n")));

/////////////////////////////////
    TCHAR  *InputFile, *OutputFile, *InvalidSwitch;   
    i = 1;
    InputFile = argv[i++];
    if(--argc) {
      if(argv[i][0] != '/') {
        OutputFile = argv[i++];
        argc--;
      }
      else
        OutputFile = InputFile;
    }
    else
      OutputFile = InputFile;
    i = GetOptions(&argv[i], argc, &InvalidSwitch);
    if(i == (DWORD)-1)
      return(_ftprintf(stderr, _T("OMF2D: Invalid option '%s'!\n"), InvalidSwitch));

/////////////////////////////////
    FILE *fi = _tfopen(InputFile, _T("rb"));
    if(fi == NULL)
      return(_ftprintf(stderr, _T("OMF2D: Cannot open input file!\n")));
    fseek(fi, 0, SEEK_END);
    DWORD fisize = ftell(fi);
    fseek(fi, 0, SEEK_SET);
    PVOID inbuf  = malloc(fisize);
    PVOID outbuf = malloc(fisize +FILESIZE_PLUS);
    fread(inbuf, 1, fisize, fi);
    fclose(fi);
    _ftprintf(stderr, _T("Converting \"%s\" to \"%s\"\n"), InputFile, OutputFile);
    PBYTE pread = ( PBYTE)inbuf;
    PBYTE pfileend = pread +fisize;
    pwrite = ( PBYTE)outbuf;

/////////////////////////////////
    NameEditor = gNameEditor;
    pLastLXDAXXRecord = pread;

    do {
      RecordLength  = *( PWORD)&pread[1];
      RecordType    = pread[0];
      switch(RecordType) {
        case OMF_COMENT : 
                          #include "OMF2D_COMENT.h"
                          break;

        case OMF_EXTDEF : 
                          #include "OMF2D_EXTDEF.h"
                          break;

        case OMF_PUBDEF :  
        case OMF_PUBD32 : 
                          #include "OMF2D_PUBDEF.h"
                          break;   

        case OMF_LNAMES : 
                          #include "OMF2D_LNAMES.h"
                          break;   

        case OMF_SEGDEF : 
                          #include "OMF2D_SEGDEF.h"
                          break;

        case OMF_SEGD32 : 
                          #include "OMF2D_SEGD32.h"
                          break;
                        
        case OMF_FIXU32 : 
                          #include "OMF2D_FIXU32.h"
                          break;   

        case OMF_LEDATA :
        case OMF_LEDA32 : 
                          #include "OMF2D_LEDAXX.h"
                          break;

        case OMF_LIDATA :
        case OMF_LIDA32 : 
                          #include "OMF2D_LIDAXX.h"
                          break;

        case OMF_COMDEF : 
                          #include "OMF2D_COMDEF.h"
                          break;

        case OMF_LEXTDEF:
        case OMF_LEXTD32: 
                          #include "OMF2D_LEXTDXX.h"
                          break;

        case OMF_LPUBDEF:
                          #include "OMF2D_LPUBDEF.h"
                          break;
        case OMF_LPUBD32: 
                          #include "OMF2D_LPUBD32.h"
                          break;

        case OMF_LCOMDEF: 
                          #include "OMF2D_LCOMDEF.h"
                          break;

        case OMF_LLNAMES: 
                          #include "OMF2D_LLNAMES.h"
                          break;

        case OMF_CEXTDEF: 
                          #include "OMF2D_CEXTDEF.h"
                          break;

        case OMF_COMDAT : 
        case OMF_COMD32 : 
                          #include "OMF2D_COMDXX.h"
                          break;

        case OMF_LINSYM :
        case OMF_LINS32 : 
                          #include "OMF2D_LINSXX.h"
                          break;

        omf2d_ignore    :
        default         : 
                          RecordLength += 3;
                          while(RecordLength--)
                            *pwrite++ = *pread++;
      }
    }
    while(pread < pfileend);

/////////////////////////////////
    fi = _tfopen(OutputFile, _T("wb+"));
    if(fi == NULL)
      return(_ftprintf(stderr, _T("OMF2D: Cannot create output file!\n")));
    fwrite(outbuf, 1, pwrite - ( PBYTE)outbuf, fi);
    fclose(fi);
//    free(inbuf);
//    free(outbuf);
    return(0);
  }
  __except(EXCEPTION_EXECUTE_HANDLER) {
    return(_ftprintf(stderr, _T("OMF2D: Invalid input file, invalid record or internal limits exceeded!\n")));
  }
}